/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2011 IMT Lucca
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Luca Anchora  <luca.anchora@imtlucca.it>
 */
#include "central-authority.h"
#include <ns3/log.h>
#include <ns3/pointer.h>
#include "enb-net-device.h"
#include <ns3/simulator.h>
#include <ns3/enb-phy.h>
#include <stdlib.h>
#include <stdio.h>
#include <ns3/random-variable.h>
#include <ns3/ue-record.h>
#include <ns3/ue-manager.h>

NS_LOG_COMPONENT_DEFINE ("CentralAuthority");

namespace ns3{

NS_OBJECT_ENSURE_REGISTERED (CentralAuthority);

typedef struct{
	int subchannelID;
	char allocated;
	char priority;
	Mac48Address user;
	int CQI;
} alloc_map_element;

struct optimum_list_element{
	Mac48Address user;
	unsigned int subc;
	int subc_state;
};

struct credit_queue_ele{
	Mac48Address user;
	int counter;
};

CentralAuthority::CentralAuthority()
{ 
	m_count_received = 0;
	m_count_eNB = 0;
}


CentralAuthority::~CentralAuthority(){ }


TypeId
CentralAuthority::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::CentralAuthority")
    .SetParent<Object> ()
    .AddConstructor<CentralAuthority> ()
    ;
  return tid;
}


void CentralAuthority::RegistereNB(Ptr<LtePhy> e)
{
	m_eNB.push_back(DynamicCast<EnbLtePhy>(e));
	m_count_eNB++;
}




/**
 * Implementation of a 'smart sharing' algorithm in a scenario with asymmetric
 * cell load. The basic idea is that none
 * of the operators want to perform worse than the non-sharing case. They have
 * an interest in sharing only if they can improve their utility (i.e., the data
 * rate). Thus a bargaining happens for the resources on which a conflict arises.
 * This implementation of the policy takes into consideration the fact that
 * the channel usage can differ from one eNB to the other and the cells may be
 * under-loaded. Moreover, fairness constraints might be imposed by each eNB,
 * thus each user reallocation activty is demanded to the respective eNB.
 */
void CentralAuthority::GetAllocationMapSmartAlgoAsymmLoad(void* prop_alloc_map)
{
	m_count_received++;
	m_maps.push_back(prop_alloc_map);

	if(m_count_received == m_count_eNB) 
	{
		int numOfIterations = 0; //number of iterations needed to obtain the final allocation maps
		alloc_map_element* allocation_map0 = 0, *allocation_map1 = 0;
		std::vector<prioSubChannel> subch_prio0 = m_eNB[0]->GetDownlinkPrioSubChannels();
		int num_subc0 = subch_prio0.size();

		allocation_map1 = (alloc_map_element*)m_maps[1];
		m_maps.pop_back();
		allocation_map0 = (alloc_map_element*)m_maps[0];
		m_maps.pop_back();
		int first_subch = allocation_map1[0].subchannelID;
		int subchLost0 = 0, subchLost1 = 0;//number of subchannels lost because of contentions
		int commonPool[2][2];

		List* realloc0 = 0, *realloc1 = 0;
		do
		{
			numOfIterations++;
			commonPool[0][0] = commonPool[0][1] = commonPool[1][0] = commonPool[1][1] = 0;
			subchLost0 = 0;
			subchLost1 = 0;
			if(realloc0)
				delete realloc0;
			if(realloc1)
				delete realloc1;
			realloc0 = realloc1 = 0;

			realloc0 = new List();
			realloc1 = new List();

			//Check the common pool for possible access conflicts
			for(int i = first_subch; i < num_subc0; i++)
			{
				int index2 = allocation_map0[i].subchannelID - first_subch;
				if(allocation_map0[i].allocated == 1)
				{//subchannel allocated. Check if it has been allocated also by the other eNB
					if(allocation_map1[index2].allocated == 1)
					{
						struct optimum_list_element* temp_conflict = (struct optimum_list_element*)malloc(sizeof(struct optimum_list_element));
						if(allocation_map1[index2].priority)
						{
							allocation_map0[i].allocated = 2;
							subchLost0++;
							commonPool[1][1]++;
							temp_conflict->user = allocation_map0[i].user;
							temp_conflict->subc = allocation_map0[i].subchannelID;
							temp_conflict->subc_state = allocation_map0[i].CQI;
							realloc0->enqueue((void*)temp_conflict);
						}else{
							allocation_map1[index2].allocated = 2;
							subchLost1++;
							commonPool[0][0]++;
							temp_conflict->user = allocation_map1[index2].user;
							temp_conflict->subc = allocation_map1[index2].subchannelID;
							temp_conflict->subc_state = allocation_map1[index2].CQI;
							realloc1->enqueue((void*)temp_conflict);
						}
					}else{
						allocation_map1[index2].allocated = 2;
						if(allocation_map0[i].priority)
							commonPool[0][0]++;
						else
							commonPool[1][0]++;
					}
				}else if(allocation_map1[index2].allocated == 1)
				{
					allocation_map0[i].allocated = 2;
					if(allocation_map0[i].priority)
						commonPool[0][1]++;
					else
						commonPool[1][1]++;
				}
			}//end for(int i = first_subch; i < num_subc0; i++); end of the check for collisions


			if(subchLost0)
			{//eNB0 has lost some contentions so it has to reschedule some users
				//Calculate the number of subchannels still free for eNB0
				int free0 = 0;
				for(int i=0; i < num_subc0; i++)
				{
					if(allocation_map0[i].allocated == 0 || allocation_map0[i].allocated == 3)
						free0++;
				}

				if(free0)//if free0==0 ==> no reallocation can happen
				{
					if(free0 >= subchLost0)
					{//There are enough free subchannels and thus there is no need to
					//kick off the other operator from its own common pool
						m_eNB[0]->ChangeAlloc(allocation_map0, realloc0);
	
					}else{
						List* smallList0 = new List();
						for(int i=0; i < (subchLost0 -free0); i++)
						{
							realloc0->dequeue();
						}
						for(int i = first_subch; i < num_subc0; i++)
						{
							if(allocation_map0[i].allocated == 2 && allocation_map0[i].priority==1)
							{
								allocation_map0[i].allocated = 0;
								smallList0->enqueue( malloc(sizeof(struct optimum_list_element)) );
							}
						}
						m_eNB[0]->ChangeAlloc(allocation_map0, realloc0);
						m_eNB[0]->ChangeAlloc(allocation_map0, smallList0);
						delete smallList0;
					}
				}//if(free0)

			}//end if(subchLost0)

			if(subchLost1)
			{//eNB1 has lost some contentions so it has to reschedule some users
				//Calculate the number of subchannels still free for eNB1
				int free1 = 0;
				for(int i=0; i < num_subc0; i++)
				{
					if(allocation_map1[i].allocated == 0 || allocation_map1[i].allocated == 3)
						free1++;
				}

				if(free1)//if free1==0 ==> no reallocation can happen
				{
					if(free1 >= subchLost1)
					{//There are enough free subchannels and thus there is no need to
					//kick off the other operator from its own common pool
						m_eNB[1]->ChangeAlloc(allocation_map1, realloc1);
	
					}else{
						List* smallList1 = new List();
						for(int i=0; i < (subchLost1 -free1); i++)
						{
							realloc1->dequeue();
						}
						for(int i = first_subch; i < num_subc0; i++)
						{
							if(allocation_map1[i].allocated == 2 && allocation_map1[i].priority==1)
							{
								allocation_map1[i].allocated = 0;
								smallList1->enqueue( malloc(sizeof(struct optimum_list_element)) );
							}
						}
						m_eNB[1]->ChangeAlloc(allocation_map1, realloc1);
						m_eNB[1]->ChangeAlloc(allocation_map1, smallList1);
						delete smallList1;
					}
				}//end if(free1)
			}//end if(subchLost1)

		}while(subchLost0 || subchLost1);

		m_eNB[0]->DoTx(allocation_map0);
		m_eNB[1]->DoTx(allocation_map1);


		m_maps.clear();
		m_count_received = 0;
	}
}


};//namespace ns3
