/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 TELEMATICS LAB, DEE - Politecnico di Bari
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Giuseppe Piro  <g.piro@poliba.it>
 */

#ifndef IP_CLASSIFIER_RECORD_H_
#define IP_CLASSIFIER_RECORD_H_

#include "ns3/ipv4-address.h"
#include "ns3/object.h"
#include "ns3/log.h"
#include <vector>
#include <stdint.h>

namespace ns3 {

/**
 * This class implements a set of parametes used to map a packet
 * into a particular bearer. This function is provided by the ip-classifier
 */
class IpClassifierRecord : public Object
{
public:
  IpClassifierRecord ();
  ~IpClassifierRecord ();
  /**
   * \brief creates a classifier records and sets all its parameters
   * \param srcAddress the source ip address
   * \param srcMask the mask to apply on the source ip address
   * \param dstAddress the destination ip address
   * \param dstMask the mask to apply on the destination ip address
   * \param srcPortLow the lower boundary of the source port range
   * \param srcPortHigh the higher boundary of the source port range
   * \param dstPortLow the lower boundary of the destination port range
   * \param dstPortHigh the higher boundary of the destination port range
   * \param protocol the L4 protocol (udp OR tcp)
   */
  IpClassifierRecord (Ipv4Address srcAddress, Ipv4Mask srcMask,
                      Ipv4Address dstAddress, Ipv4Mask dstMask,
                      uint16_t srcPortLow, uint16_t srcPortHigh,
                      uint16_t dstPortLow, uint16_t dstPortHigh,
                      uint8_t protocol);

  /**
   * \brief add a new source ip address to the classifier
   * \param srcAddress the source ip address
   * \param srcMask the mask to apply on the source ip address
   */
  void AddSrcAddr (Ipv4Address srcAddress, Ipv4Mask srcMask);
  /**
    * \brief add a new destination ip address to the classifier
    * \param dstAddress the destination ip address
    * \param dstMask the mask to apply on the destination ip address
    */
  void AddDstAddr (Ipv4Address dstAddress, Ipv4Mask dstMask);
  /**
   * \brief add a range of source port to the classifier
   * \param srcPortLow the lower boundary of the source port range
   * \param srcPortHigh the higher boundary of the source port range
   */
  void AddSrcPortRange ( uint16_t srcPortLow, uint16_t srcPortHigh);
  /**
   * \brief add a range of destination port to the classifier
   * \param dstPortLow the lower boundary of the destination port range
   * \param dstPortHigh the higher boundary of the destination port range
   */
  void AddDstPortRange ( uint16_t dstPortLow, uint16_t dstPortHigh);
  /**
   * \brief add a protocol to the classifier
   * \param proto the L4 protocol to add
   */
  void AddProtocol (uint8_t proto);


  /**
   * \brief check if a packets can be used with this classifier
   * \param srcAddress the source ip address of the packet
   * \param dstAddress the destination ip address of the packet
   * \param srcPort the source port of the packet
   * \param dstPort the destination port of the packet
   * \param proto The L4 protocol of the packet
   */
  bool CheckMatch (Ipv4Address srcAddress,
                   Ipv4Address dstAddress,
                   uint16_t srcPort,
                   uint16_t dstPort,
                   uint8_t proto) const;


private:

  struct PortRange
  {
    uint16_t PortLow;
    uint16_t PortHigh;
  };
  struct ipv4Addr
  {
    Ipv4Address Address;
    Ipv4Mask Mask;
  };


  std::vector<uint8_t> m_protocol;
  std::vector<struct ipv4Addr> m_srcAddr;
  std::vector<struct ipv4Addr> m_dstAddr;
  std::vector<struct PortRange> m_srcPortRange;
  std::vector<struct PortRange> m_dstPortRange;

};

}
; // namespace ns3

#endif /* IP_CLASSIFIER_RECORD_H_ */
