/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 TELEMATICS LAB, DEE - Politecnico di Bari
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Original Author: Giuseppe Piro  <g.piro@poliba.it>
 * Modified by: Francesco Guidolin <francesco.guidolin@gmail.com> 
 */

#include <cmath>
#include <ns3/log.h>
#include "jakes-fading-loss-model.h"
#include <ns3/simulator.h>
#include <stdint.h>
#include "stdlib.h"
#include "lte-phy.h"
#include <ns3/mobility-model.h>


NS_LOG_COMPONENT_DEFINE ("JakesFadingLossModel");

namespace ns3 {
	
	
	NS_OBJECT_ENSURE_REGISTERED (JakesFadingLossModel);
	
	JakesFadingLossModel::JakesFadingLossModel ()
	: m_nbOfPaths (1, 4),
    m_startJakes (1, 2000),
    m_phy (0),
    m_betaO(0),
    m_m(0)    
	{
		NS_LOG_FUNCTION (this);
		SetSamplingPeriod (0.1); // default value -> update scatterers number
	}
	
	
	TypeId
	JakesFadingLossModel::GetTypeId (void)
	{
		static TypeId tid = TypeId ("ns3::JakesFadingLossModel")
		.SetParent<DiscreteTimeLossModel> ()
		.AddConstructor<JakesFadingLossModel> ()
		;
		return tid;
	}
	
	
	JakesFadingLossModel::~JakesFadingLossModel ()
	{
		m_phy  = 0;
	}
	
	
	void
	JakesFadingLossModel::SetPhy (Ptr<LtePhy> phy)
	{
		NS_LOG_FUNCTION (this);
		m_phy = phy;
	}
	
	
	Ptr<LtePhy>
	JakesFadingLossModel::GetPhy (void)
	{
		NS_LOG_FUNCTION (this);
		return m_phy;
	}
	
	
	// Added by Francesco Guidolin	
	double
	JakesFadingLossModel::GetValue (double frequency)
	{		
		
		if (NeedForUpdate ())
		{
		        	
			// compute number of path m
			// x = 1 -> M=6, x = 2 -> M=8, x = 3 -> M=10, x = 4 -> M=12
			int x = m_nbOfPaths.GetValue ();
			
			if (x == 1)
				m_m = 6;
			else if (x == 2)
				m_m=8;
			else if (x == 3)
				m_m = 10;
			else if (x == 4)
				m_m = 12; 
                       //NS_LOG_UNCOND ("Update m " << m_m );  
                
                SetLastUpdate ();           
		} 
                  


		double now = Simulator::Now ().GetSeconds ();
		
		double multipath = JakesFadingLossModel::CalculateValue (frequency,now,m_m);
		
		return multipath;
	}
	
	
	
	double JakesFadingLossModel::CalculateValue (double frequency, double now, int m)
	
	{         
		
		Ptr<MobilityModel> mobility = 0;
		
		
		// this needs to be fixed, we cannot allow a propagation model to need pointers to all PHYs
		mobility = GetPhy ()->GetDownlinkSpectrumPhy ()->GetMobility ()->GetObject<MobilityModel> ();
		Vector speedVector = mobility->GetVelocity ();
		
		
		
		double speed = sqrt (pow (speedVector.x,2) +  pow (speedVector.y,2));
	
		
		int n =2*((2*m)+1);
		double wc = 2* PI *frequency; // pulsation
		double wm = wc *speed/C ; // max doppler shift
		
		//NS_LOG_UNCOND ("SPEED: " << speed);
		//NS_LOG_UNCOND ("Frequency under analysis: " << frequency);
		//NS_LOG_UNCOND ("Frequency Doppler shift: " << wm/(2*PI));
		
		double hI=0;
		double hQ=0;
		double hIsum = 0;
		double hQsum = 0;
		double hInonsum = 0;
		double hQnonsum = 0;
		double betaN = 0;
                  
		
                
		int count;
		for (count = 1 ; count < (m+1); count++ )
		{ 
			betaN = (PI*count)/(m+1); 
			hIsum = hIsum + (cos(betaN)*cos(cos(2*PI*count/n)*wm*now));                           
		}
		
		hIsum = hIsum*2;
		hInonsum = sqrt(2)*cos(wm*now)*cos(m_betaO);     
		

                hI = (hIsum + hInonsum)*(1/(sqrt(2*m+1)));
		                
                		
		
		for (count = 1; count < (m+1); count++ )
		{
			betaN = (PI*count)/(m+1); 
			hQsum = hQsum + (sin(betaN)*cos(cos(2*PI*count/n)*wm*now));
		}
		
		hQsum = hQsum*2;
		hQnonsum = sqrt(2)*cos(wm*now)*sin(m_betaO); 
                
		hQ = (hQsum + hQnonsum) * (1/(sqrt(2*m+1)));		
               		
		
		double multipath = sqrt(pow(hI,2)+pow(hQ,2)); 
		double multipath_dB = 10*log10(multipath);
		             
		//NS_LOG_UNCOND ("Associated MULTIPATH value : " << multipath_dB << " [dB]");
		
		return multipath_dB;
	}
	
	
	
	
} // namespace ns3
