/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2011 IMT Lucca
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Luca Anchora  <luca.anchora@imtlucca.it>
 */


#include "multiOp-packet-scheduler.h"
#include <ns3/log.h>
#include <ns3/pointer.h>
#include "ue-net-device.h"
#include "enb-net-device.h"
#include "enb-net-device.h"
#include "rlc-entity.h"
#include "rrc-entity.h"
#include "lte-mac-header.h"
#include "enb-mac-entity.h"
#include "ue-manager.h"
#include "ue-record.h"
#include "amc-module.h"
#include <ns3/random-variable.h>
#include <math.h>
#include <ns3/simulator.h>
#include "List.h" /* Luca Anchora */
#include "enb-phy.h"
#include <ns3/central-authority.h> /* Luca Anchora*/
#include <stdio.h> /* Luca Anchora */
#include <stdlib.h> /* Luca Anchora */

NS_LOG_COMPONENT_DEFINE ("MultiOpPacketScheduler");

namespace ns3 {


NS_OBJECT_ENSURE_REGISTERED (MultiOpPacketScheduler);

#define PACKET_SIZE 1000 /*bytes*/

MultiOpPacketScheduler::MultiOpPacketScheduler ()
{
	NS_LOG_FUNCTION (this);
	SetDevice (0);
	m_queues = 0;
	m_totThroughput = 0;
	m_currThroughput = 0;
	m_totCapacity = 0;
	m_currCapacity = 0;
	m_countTx = 0;
	prop_alloc_map = 0;
	alloc_map_noshar = 0;
	optimum_list_ = 0;
	optimum_list_noshar_ = 0;
	m_noSharCap = 0;
	m_noSharThroughput = 0;
	m_initialSharCap = 0;
	m_initialSharThroughput = 0;
	m_unallocatedSlots = 0;
	m_commonSlotsUsed = 0;
}

TypeId
MultiOpPacketScheduler::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MultiOpPacketScheduler")
    .SetParent<PacketScheduler> ()
    .AddConstructor<MultiOpPacketScheduler> ()
  ;
  return tid;
}


/**
 * In the destructor we deallocate all the pointers that have been
 * dynamically allocated, in order to avoid memory leaks.
 */
MultiOpPacketScheduler::~MultiOpPacketScheduler ()
{
	if(m_queues)
		free(m_queues);
	m_queues = 0;
	if(optimum_list_)
	{
		optimum_list_->start_scanning();
		struct optimum_list_element* ele = (struct optimum_list_element*)optimum_list_->next_item();
		while(ele)
		{
			free(ele);
			ele = 0;
			ele = (struct optimum_list_element*)optimum_list_->next_item();
		}
		delete(optimum_list_);
	}
	optimum_list_ = 0;

	optimum_list_noshar_ = 0;

	if(prop_alloc_map)
		free(prop_alloc_map);
	prop_alloc_map = 0;

	if(alloc_map_noshar)
		free(alloc_map_noshar);
	alloc_map_noshar = 0;

}


/**
 * Allocate a flow queue per UE. The current implementation does not address the case
 * of a change in the number of UEs. Indeed, when a UE unregisters from the eNB or
 * a new one comes, then the corresponding queue should be updated accordingly.
*/
void
MultiOpPacketScheduler::allocate_queues(int num_ue, Ptr<EnbNetDevice> enb)
{
	m_queues = (prioQueue*) malloc (sizeof(prioQueue)*num_ue);
	if(!m_queues)
	{
		printf("Not enough space for allocating flow queues in the scheduler. Terminating the process\n");
		exit(1);
	}
	std::vector<Ptr<UeRecord> >* listUE;
	listUE = enb->GetUeManager ()-> GetUeRecords ();
	for(int i=0; i<num_ue; i++)
	{
		m_queues[i].priority = 0; //maximum priority to all flows as a default setting
		m_queues[i].credit_counter = 0.0;
		m_queues[i].pkt_queue = new List();
		m_queues[i].tot_byte_tx = 0.0;
		m_queues[i].destUE = Mac48Address::ConvertFrom(listUE-> at(i)-> GetUe ()-> GetAddress ());
	}

}


/**
 *  Building of the ordered list of tuples <user, subchannel, subchann_state>. The list is 
 *  ordered with respect to the value of the subchannel state. In this way, we have a
 *  clear and ordered list of the channel state for each user that can be exploited
 *  by the resource allocator for the transmission of packets on the best subchannels.
 */
void MultiOpPacketScheduler::build_optimum_list(void)
{
	
	List *new_optimum_list_ = 0;
	new_optimum_list_ = new List();
	if(new_optimum_list_==0)
	{
		std::cout << "Error: not enough memory to allocate optimum_list_" << std::endl;
		exit(1);
	}
	struct optimum_list_element* element = 0;
	struct optimum_list_element* tmp = 0;
	int position_counter = 0; //needed to put the list element in the correct position
	int totSubChannels = 0;
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	Ptr<UeRecord> ueRecord;
	UeRecord::CqiFeedbacks feedbacks;
	std::vector<prioSubChannel> prioSubchann = enb->GetPhy()->GetDownlinkPrioSubChannels();
	int firstSubc = prioSubchann[0].subChannel;

	int num_privSubchannels = 0;
	for(unsigned int i=0; i<prioSubchann.size(); i++)
	{
		if(prioSubchann[i].priority)
			num_privSubchannels++;
	}


	std::vector<Ptr<UeRecord> >* it_recs = enb->GetUeManager ()->GetUeRecords ();
	for (std::vector<Ptr<UeRecord> >::iterator iter = it_recs->begin (); iter != it_recs->end (); ++iter)
	{
		feedbacks = (*iter)->GetCqiFeedbacks ();
		totSubChannels = feedbacks.size ();
		int inf_priv = 0, sup_priv = 0;
		for (int i = 0; i < totSubChannels; i++)
		{
			int subChannel = feedbacks.at (i).m_subChannelId;
			int cqi = feedbacks.at (i).m_cqi;

			element = (struct optimum_list_element*)malloc(sizeof(struct optimum_list_element));
			element->user = Mac48Address::ConvertFrom ((*iter)->GetUe ()->GetAddress ());
			element->subc = subChannel;
			element->subc_state = cqi;

			//Insert the element in the correct position into the ordered list according to
			//the value of the subchannel state. The list is scanned and when the correct
			//position is found, the element is inserted.
			//Elements are put in decreasing order of CQI.
			new_optimum_list_->start_scanning();
			tmp = 0;
			tmp = (struct optimum_list_element*)new_optimum_list_->next_item();
			position_counter = 1;
			char priv_tmp = 0;
			char pub_tmp = 0;
			char priv_ele = 0;
			char pub_ele = 0;
			int sharedSubc = totSubChannels - num_privSubchannels;
			if(!firstSubc && prioSubchann[0].priority)
			{
				inf_priv = firstSubc;
				sup_priv = num_privSubchannels - sharedSubc - 1;
			}else{
				inf_priv = firstSubc + 2*sharedSubc;
				sup_priv = inf_priv - sharedSubc + num_privSubchannels - 1;
			}

			if(((int)element->subc >= inf_priv) && ((int)element->subc <= sup_priv))
				priv_ele = 1;
			else
				priv_ele = 0;

			pub_ele = 0;
			if(!priv_ele)
			{
				if(!firstSubc && prioSubchann[0].priority)
				{
					if((int)element->subc <= (sup_priv + sharedSubc))
						pub_ele = 1;
				}else{
					if((int)element->subc >= (inf_priv - sharedSubc))
						pub_ele = 1;
				}
			}

			while(tmp!=0)
			{
				if(((int)tmp->subc >= inf_priv) && ((int)tmp->subc <= sup_priv))
					priv_tmp = 1;
				else
					priv_tmp = 0;

				pub_tmp = 0;
				if(!priv_tmp)
				{
					if(!firstSubc && prioSubchann[0].priority)
					{
						if((int)tmp->subc <= (sup_priv + sharedSubc))
							pub_tmp = 1;
					}else{
						if((int)tmp->subc >= (inf_priv - sharedSubc))
							pub_tmp = 1;
					}
				}

				if(tmp->subc_state < element->subc_state)
				{
					break;
				}else if((tmp->subc_state == element->subc_state) && priv_ele && !priv_tmp)
				{
					break;
				}else if((tmp->subc_state == element->subc_state) && priv_ele && priv_tmp)
				{
					UniformVariable* u = new UniformVariable();
					if(u->GetValue(0.0, 1.0) > 0.5)
						break;
					delete u;
				}else if((tmp->subc_state == element->subc_state) && (pub_ele && !pub_tmp))
				{
					break;
				}else if((tmp->subc_state == element->subc_state) && (pub_ele && pub_tmp))
				{
					UniformVariable* u = new UniformVariable();
					if((pub_tmp && u->GetValue(0.0, 1.0) > 0.5))
						break;
					delete u;
				}else{
					tmp = (struct optimum_list_element*)new_optimum_list_->next_item();
					position_counter++;
				}
			}//end while(tmp!=0)
			new_optimum_list_->put_position(element, position_counter);

		} //end for(subc=0; subc<tot_subc; subc++)
	} //end for (std::vector<UeRecord*>::iterator iter = it_recs->begin (); iter != it_recs->end (); ++iter)

	if(new_optimum_list_->len())
		optimum_list_ = new_optimum_list_;

	return;
}


void MultiOpPacketScheduler::build_optimum_list_noshar(void)
{
	
	List *new_optimum_list_noshar_ = 0;
	new_optimum_list_noshar_ = new List();
	if(new_optimum_list_noshar_ == 0)
	{
		std::cout << "Error: not enough memory to allocate optimum_list_noshar_" << std::endl;
		exit(1);
	}
	struct optimum_list_element* element = 0;
	struct optimum_list_element* tmp = 0;
	int position_counter = 0; //needed to put the list element in the correct position
	int totSubChannels = 0;
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	Ptr<UeRecord> ueRecord;
	UeRecord::CqiFeedbacks feedbacks;
	std::vector<prioSubChannel> prioSubchann = enb->GetPhy()->GetDownlinkPrioSubChannels();
	int firstSubc = prioSubchann[0].subChannel;
	int num_privSubchannels = 0;
	for(unsigned int i=0; i<prioSubchann.size(); i++)
	{
		if(prioSubchann[i].priority)
			num_privSubchannels++;
	}

	std::vector<Ptr<UeRecord> >* it_recs = enb->GetUeManager ()->GetUeRecords ();
	for (std::vector<Ptr<UeRecord> >::iterator iter = it_recs->begin (); iter != it_recs->end (); ++iter)
	{
		feedbacks = (*iter)->GetCqiFeedbacks ();
		totSubChannels = feedbacks.size ();

		for (int i = 0; i < totSubChannels; i++)
		{
			int subChannel = feedbacks.at (i).m_subChannelId;
			int cqi = feedbacks.at (i).m_cqi;
			int index = subChannel - firstSubc;
			if(prioSubchann[index].priority == 0)//skip the subchannels belonging to the other operators
				continue;

			element = (struct optimum_list_element*)malloc(sizeof(struct optimum_list_element));
			element->user = Mac48Address::ConvertFrom ((*iter)->GetUe ()->GetAddress ());
			element->subc = subChannel;
			element->subc_state = cqi;

			//Insert the element in the correct position into the ordered list according to
			//the value of the subchannel state. The list is scanned and when the correct
			//position is found, the element is inserted.
			//Elements are put in decreasing order of CQI.
			new_optimum_list_noshar_->start_scanning();
			tmp = 0;
			tmp = (struct optimum_list_element*)new_optimum_list_noshar_->next_item();
			position_counter = 1;
			while(tmp!=0)
			{
				if(tmp->subc_state < element->subc_state)
				{
					break;
				}else if (tmp->subc_state == element->subc_state)
				{
					UniformVariable* u = new UniformVariable();
					double val = u->GetValue(0.0, 1.0);
					delete u;
					if(val > 0.5)
					{
						tmp = (struct optimum_list_element*)new_optimum_list_noshar_->next_item();
						position_counter++;
					}else{
						break;
					}
				}else{
					tmp = (struct optimum_list_element*)new_optimum_list_noshar_->next_item();
					position_counter++;
				}
			}//end while(tmp!=0)

			new_optimum_list_noshar_->put_position(element, position_counter);

		} //end for(subc=0; subc<tot_subc; subc++)
	} //end for (std::vector<UeRecord*>::iterator iter = it_recs->begin (); iter != it_recs->end (); ++iter)

	if(new_optimum_list_noshar_->len())
		optimum_list_noshar_ = new_optimum_list_noshar_;

	return;
}

void printMap(alloc_map_element* map, int len)
{
	printf("Subch\tUser\tCQI\tprio\talloc\n");
	for(int i=0; i < len; i++)
	{
		printf("%d\t", map[i].subchannelID);
		std::cout << map[i].user <<"\t";
		printf("%d\t", map[i].CQI);
		printf("%d\t", map[i].priority);
		printf("%d\n", map[i].allocated);
	}
}


void MultiOpPacketScheduler::printChannelState()
{
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	Ptr<UeRecord> ueRecord;
	UeRecord::CqiFeedbacks feedbacks;

	uint8_t buffer[6];
	memset(buffer, 0, 6);
	Mac48Address currAddr;
	
	std::vector<Ptr<UeRecord> >* it_recs = enb->GetUeManager ()->GetUeRecords ();
	for (std::vector<Ptr<UeRecord> >::iterator iter = it_recs->begin (); iter != it_recs->end (); ++iter)
	{
		feedbacks = (*iter)->GetCqiFeedbacks ();
		int totSubChannels = feedbacks.size ();

		for (int i = 0; i < totSubChannels; i++)
		{
			int subChannel = feedbacks.at (i).m_subChannelId;
			int cqi = feedbacks.at (i).m_cqi;
			currAddr = Mac48Address::ConvertFrom ((*iter)->GetUe ()->GetAddress ());
			currAddr.CopyTo(buffer);

			uint8_t UEId = buffer[5];
			char name[255];
			memset(name, 0, 255);
			sprintf(name, "logUE%u_channel.txt", UEId);
			FILE *fp = fopen(name, "a");
			double now = Simulator::Now().GetSeconds();
			fprintf(fp, "%f\t%d\t%d\n", now, subChannel, cqi);

			fflush(fp);
			fclose(fp);
		}
	}
	
}


/**
 * Implementation of a possible Packet Scheduler (PS) plus a 
 * Resource Allocator (RA).
 */
void
MultiOpPacketScheduler::DoRunPacketScheduler (void)
{
	NS_LOG_FUNCTION (this);

	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	int num_ue = enb->GetUeManager ()->GetNRegisteredUes();

	if(!m_queues)
	{ //it is the first transmission, so a virtual packet is sent to obtain info 
	  //about the channel state
		TxVirtualPacket();
		allocate_queues(num_ue, enb);
		return;
	}

	build_optimum_list();//build the list needed by the resource allocator to map packets onto subchannels
	build_optimum_list_noshar();//optimum list but with spectrum sharing = 0%
	
	//Fill in the transmission queue by picking packets from
	//the flow queues. The number of packets taken from each queue
	//depends on the number of resources available (frequency subchannels)
	//and on the scheduling policy. In this case we do not care about fairness.
	//Packets are selected according to a greedy max throughput criterion.
	//We suppose that all the flows are always backlogged!
	std::vector<prioSubChannel> prioSubchann = enb->GetPhy()->GetDownlinkPrioSubChannels();
	int priv_resources = 0;
	int tot_resources = prioSubchann.size();
	for(int i=0; i<tot_resources; i++)
	{
		if(prioSubchann[i].priority)
			priv_resources++;
	}
	int res_allocated = 0;
	int first_res = prioSubchann[0].subChannel;



	int privResnoShar = priv_resources;
	priv_resources = tot_resources;

	//Creation of the allocation map proposed by the eNB considering the common pool
	if(prop_alloc_map)
		free(prop_alloc_map);
	prop_alloc_map = 0;
	prop_alloc_map = (alloc_map_element*)malloc(sizeof(alloc_map_element) * tot_resources);
	if(!prop_alloc_map)
	{
		std::cout<<"Error: not enough memory to allocate the proposed allocation map.\n";
		exit(1);
	}
	//Creation of the allocation map proposed by the eNB without considering the common pool
	if(alloc_map_noshar)
		free(alloc_map_noshar);
	alloc_map_noshar = 0;
	alloc_map_noshar = (alloc_map_element*)malloc(sizeof(alloc_map_element) * tot_resources);
	if(!alloc_map_noshar)
	{
		std::cout<<"Error: not enough memory to allocate the proposed allocation map noshar.\n";
		exit(1);
	}

	//Initialization of the allocation maps
	for (int i=0; i < tot_resources; i++)
	{
		prop_alloc_map[i].subchannelID = prioSubchann[i].subChannel;
		prop_alloc_map[i].priority = prioSubchann[i].priority;
		prop_alloc_map[i].allocated = 0;
		alloc_map_noshar[i].subchannelID = prioSubchann[i].subChannel;
		alloc_map_noshar[i].priority = prioSubchann[i].priority;
		alloc_map_noshar[i].allocated = 0;
	}

	int copyCounter[num_ue];
	for(int i=0; i<num_ue; i++)
	{
		m_queues[i].credit_counter = 0;//necessary for a possible credit-based packet scheduler
		copyCounter[i] = 0;
	}

	res_allocated = 0;
	optimum_list_noshar_->start_scanning();
	struct optimum_list_element* tmp = 0;
	m_noSharCap = m_initialSharCap = 0.0;
	m_noSharThroughput = m_initialSharThroughput = 0.0;
	
	//By setting two thresholds we are sure that all the UEs will receive at least 1 resource
	int count_inf = 0;
	int infTH = 2; //priv_resources/num_ue;
	int supTH = 2; //priv_resources/num_ue + 1;
	while(res_allocated < privResnoShar && res_allocated < (2*num_ue))
	{
		int queues_idx = 0;
		bool fairnessCond = 1;
		if(count_inf == num_ue)//restarts from the beginning of the list
			optimum_list_noshar_->start_scanning();

		do
		{
			tmp = (struct optimum_list_element*)optimum_list_noshar_->next_item();
			if(tmp)
			{
				for(queues_idx=0; queues_idx < num_ue; queues_idx++)
				{
					if(m_queues[queues_idx].destUE == tmp->user)
						break;
				}
			}
			fairnessCond = (((copyCounter[queues_idx] >=infTH) && (count_inf < num_ue)) || (copyCounter[queues_idx] > supTH));
		}while(tmp && (alloc_map_noshar[tmp->subc - first_res].allocated!=0 || fairnessCond));

		if(tmp)
		{
			int idx = tmp->subc - first_res;
			alloc_map_noshar[idx].allocated = 1;
			alloc_map_noshar[idx].user = tmp->user;
			alloc_map_noshar[idx].CQI = tmp->subc_state;
			res_allocated++;

			copyCounter[queues_idx]++;
			if(copyCounter[queues_idx] >= infTH)
				count_inf++;

		}else{
			std::cout << "Error: optimum_list_noshar finished but allocation map not completed\n";
			exit(1);
		}
	}


	res_allocated = 0;
	tmp = 0;
	count_inf = 0;
	//Allocate all the available resources (subchannels) to the best users
	optimum_list_->start_scanning();

	while(res_allocated < tot_resources && res_allocated < (2*num_ue))
	{
		int queues_idx = 0;
		bool fairnessCond = 1;
		if(count_inf == num_ue) //go back to the beginning of the list
			optimum_list_->start_scanning();

		do
		{
			tmp = (struct optimum_list_element*)optimum_list_->next_item();
			if(tmp)
			{
				for(queues_idx=0; queues_idx < num_ue; queues_idx++)
				{
					if(m_queues[queues_idx].destUE == tmp->user)
						break;
				}
			}
			fairnessCond = ( ( (m_queues[queues_idx].credit_counter >=infTH) && (count_inf < num_ue)) ||(m_queues[queues_idx].credit_counter >= supTH) );
		}while(tmp && (prop_alloc_map[tmp->subc - first_res].allocated!=0 || fairnessCond));

		if(tmp)
		{
			int idx = tmp->subc - first_res;
			prop_alloc_map[idx].user = tmp->user;
			prop_alloc_map[idx].CQI = tmp->subc_state;
			if(res_allocated < priv_resources)
			{
				prop_alloc_map[idx].allocated = 1;
				m_queues[queues_idx].credit_counter++;
				if(m_queues[queues_idx].credit_counter >= infTH)
					count_inf++;
			}else{
				prop_alloc_map[idx].allocated = 3;//the subchannel is not allocated, but a 
				                                  //preference about the user is expressed
			}
			res_allocated++;
		}else{
			std::cout << "Error: optimum_list_ finished but allocation map not completed\n";
			exit(1);
		}
	}


	(DynamicCast<EnbLtePhy>(enb->GetPhy()))->GetCA()->GetAllocationMapSmartAlgoAsymmLoad(prop_alloc_map);
}


/**
 * The eNB sends a virtual packet to allow its UEs to compute the SINR
 * on each subchannel used for the downlink.
*/
void
MultiOpPacketScheduler::TxVirtualPacket ()
{
	NS_LOG_FUNCTION (this);
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	
	Ptr<PacketBurst> pb = CreateObject<PacketBurst> ();
	Ptr<Packet> p = new Packet (1);
	
	Mac48Address from = Mac48Address::ConvertFrom (enb->GetAddress ());
	Mac48Address to = Mac48Address::ConvertFrom (enb->GetAddress ());


	LteMacHeader header;
	header.SetSource (from);
	header.SetDestination (to);
	p->AddHeader (header);
	
	pb->AddPacket (p);
	enb->SetPacketToSend (pb);
	enb->StartTransmission ();

}


/**
 * Change the proposed channel allocation map according to the map passed as argument.
 * This one indicates, for each subchannel, if it is
 * free or not. The allocation is still greedy. The fairness constraints must be
 * respected also by the new allocation.
 */
void* MultiOpPacketScheduler::NewAlloc(void* e, List* unallocated)
{
	NS_LOG_FUNCTION (this);

	alloc_map_element* map = (alloc_map_element*)e;
	unallocated->start_scanning();
	int num_unallocated = unallocated->len();
	int first_subch = map[0].subchannelID;
	struct optimum_list_element* tmp = 0;
	optimum_list_->start_scanning();

	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	std::vector<prioSubChannel> prioSubchann = enb->GetPhy()->GetDownlinkPrioSubChannels();
	int tot_resources = prioSubchann.size();
	int num_ue = enb->GetUeManager ()->GetNRegisteredUes();
	int count_inf = 0;
	int infTH = 2; //priv_resources/num_ue;
	int supTH = 2; //priv_resources/num_ue + 1;
	int res_allocated = 0;

	for(int i=0; i<num_ue; i++)
	{
		m_queues[i].credit_counter = 0;//necessary for a possible credit-based packet scheduler
	}

	//Calculate the number of packets allocated for each flow. This is needed
	//for the fairness constraints.
	for(int i=0; i<tot_resources; i++)
	{
		if(map[i].allocated==1)
		{
			for(int queues_idx=0; queues_idx < num_ue; queues_idx++)
			{
				if(m_queues[queues_idx].destUE == map[i].user)
				{
					m_queues[queues_idx].credit_counter++;
					if(m_queues[queues_idx].credit_counter >= infTH)
						count_inf++;
				}
			}
		}
	}

	while(res_allocated < num_unallocated)
	{
		int queues_idx = 0;
		bool fairnessCond = 1;
		if(count_inf == num_ue) //go back to the beginning of the list
			optimum_list_->start_scanning();

		do
		{
			tmp = (struct optimum_list_element*)optimum_list_->next_item();
			if(tmp)
			{
				for(queues_idx=0; queues_idx < num_ue; queues_idx++)
				{
					if(m_queues[queues_idx].destUE == tmp->user)
						break;
				}
			}
			fairnessCond = ( ( (m_queues[queues_idx].credit_counter >=infTH) && (count_inf < num_ue)) ||(m_queues[queues_idx].credit_counter >= supTH) );
		}while(tmp && (map[tmp->subc - first_subch].allocated!=0 || fairnessCond));

		if(tmp)
		{
			int idx = tmp->subc - first_subch;
			map[idx].user = tmp->user;
			map[idx].CQI = tmp->subc_state;
			map[idx].allocated = 1;
			m_queues[queues_idx].credit_counter++;
			if(m_queues[queues_idx].credit_counter >= infTH)
				count_inf++;
			
			res_allocated++;
		}else{
			break; //no more free resources
		}
	}


	return map;
}


/**
 * Execute the transmission on the basis of the resource allocation indications
 * given in the map. In the current implementation, only one packet is transmitted
 * for each subchannel and its dimension depends on the achievable capacity.
 */
void MultiOpPacketScheduler::DoTx(void* e)
{
	NS_LOG_FUNCTION (this);
	alloc_map_element* map = (alloc_map_element*)e;
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	Ptr<PacketBurst> pb = CreateObject<PacketBurst> ();
	Ptr<Packet> p;
	LteMacHeader header;
	Ptr<UeRecord> ueRecord;
	Mac48Address from = Mac48Address::ConvertFrom (enb->GetAddress ());//MAC address of this eNB node,
	                                                                   //that is the sender of the packets
	Mac48Address destinationMac;
	Ptr<PdcchMapIdealControlMessage> msg = CreateObject<PdcchMapIdealControlMessage> ();//control packet
	
	std::vector<prioSubChannel> prioSubchann = enb->GetPhy()->GetDownlinkPrioSubChannels();
	int initial_subch = 0;
	int tot_res = prioSubchann.size();
	for(int i=0; i<tot_res; i++)
	{
		if(prioSubchann[i].priority)
			initial_subch++;
	}


	int count_allocated = 0;//number of resources allocated


	for(int i=0; i < tot_res; i++)
	{
		if(map[i].allocated==1)
		{//Subchannel allocated to this eNB. A packet is scheduled to be transmitted
		 //on it.
	
			int mcs;
			destinationMac = map[i].user;		
			ueRecord = enb->GetUeManager ()->GetUeRecord (destinationMac);
			
			mcs = GetMacEntity ()->GetAmcModule ()->GetMCSFromCQI (map[i].CQI);
			msg->AddNewRecord (PdcchMapIdealControlMessage::DOWNLINK,
					   map[i].subchannelID,
					   ueRecord->GetUe ()->GetObject<UeNetDevice> (),
					   mcs);
			
			
						
			p = new Packet(PACKET_SIZE);
			header.SetSource (from);
			header.SetDestination (destinationMac);
			p->AddHeader (header);
			pb->AddPacket (p);
			count_allocated++;

			if(map[i].priority == 0)
				m_commonSlotsUsed++;

		}
		
	}

	m_unallocatedSlots = initial_subch - count_allocated;

	enb->SetPacketToSend (pb);
	GetMacEntity ()->GetObject<EnbMacEntity> ()->SendPdcchMapIdealControlMessage (msg);
	enb->StartTransmission ();
}


};//namespace
