/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 TELEMATICS LAB, DEE - Politecnico di Bari
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Giuseppe Piro  <g.piro@poliba.it>
 */

#include <cmath>
#include <ns3/log.h>
#include "multipath-loss-model.h"
#include <ns3/simulator.h>
#include <stdint.h>
#include "stdlib.h"
#include "lte-phy.h"
#include <ns3/mobility-model.h>

#include "JakesTraces/multipath_v0_M6.h"
#include "JakesTraces/multipath_v0_M8.h"
#include "JakesTraces/multipath_v0_M10.h"
#include "JakesTraces/multipath_v0_M12.h"

#include "JakesTraces/multipath_v3_M6.h"
#include "JakesTraces/multipath_v3_M8.h"
#include "JakesTraces/multipath_v3_M10.h"
#include "JakesTraces/multipath_v3_M12.h"

#include "JakesTraces/multipath_v30_M6.h"
#include "JakesTraces/multipath_v30_M8.h"
#include "JakesTraces/multipath_v30_M10.h"
#include "JakesTraces/multipath_v30_M12.h"

#include "JakesTraces/multipath_v120_M6.h"
#include "JakesTraces/multipath_v120_M8.h"
#include "JakesTraces/multipath_v120_M10.h"
#include "JakesTraces/multipath_v120_M12.h"


NS_LOG_COMPONENT_DEFINE ("MultipathLossModel");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (MultipathLossModel);

MultipathLossModel::MultipathLossModel ()
{
  NS_LOG_FUNCTION (this);
//  SetSamplingPeriod (0.5); // default value. Commented by Luca Anchora
SetSamplingPeriod (0.001); //Luca Anchora:1 msec
  m_multipath = new MultipathForFrequencyDomain ();
  m_nbOfPaths = new UniformVariable (1, 4);
  m_startJakes = new UniformVariable (1, 2000);
  m_phy = 0;
  m_betaO = 0; //Luca Anchora

//BEGIN Luca Anchora
	int x = m_nbOfPaths->GetValue ();
			
	if (x == 1)
		m_m = 6;
	else if (x == 2)
		m_m=8;
	else if (x == 3)
		m_m = 10;
	else if (x == 4)
		m_m = 12; 
//END Luca Anchora
}


TypeId
MultipathLossModel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::MultipathLossModel")
    .SetParent<LossModel> ()
    .AddConstructor<MultipathLossModel> ()
  ;
  return tid;
}


MultipathLossModel::~MultipathLossModel ()
{
  m_multipath = 0;
  m_nbOfPaths = 0;
  m_startJakes = 0;
}


void
MultipathLossModel::SetPhy (Ptr<LtePhy> phy)
{
  NS_LOG_FUNCTION (this);
  m_phy = phy;

  SetValue ();
}


Ptr<LtePhy>
MultipathLossModel::GetPhy (void)
{
  NS_LOG_FUNCTION (this);
  return m_phy;
}


void
MultipathLossModel::SetValue (void)
{
  NS_LOG_FUNCTION (this);

  m_multipath->clear ();

  int downlinkSubChannels = GetPhy ()->GetDownlinkSubChannels ().size ();

  Ptr<MobilityModel> mobility = GetPhy ()->GetDownlinkSpectrumPhy ()->GetMobility ()->GetObject<MobilityModel> ();
  Vector speedVector = mobility->GetVelocity ();

  double speed = sqrt (pow (speedVector.x,2) +  pow (speedVector.y,2));

  NS_LOG_FUNCTION (this << mobility << speedVector << speed);

  if (speed < 3.)
    {
      speed = 0;
    }
  else if (speed < 30.)
    {
      speed = 3.;
    }
  else if (speed < 120.)
    {
      speed = 30.;
    }
  else
    {
      speed = 30;
    }

  NS_LOG_FUNCTION (this << mobility << speedVector << speed);


  /*
   * Jackes Model.
   * Jakes popularised a model for Rayleigh fading based on summing sinusoids
   * William C. Jakes, Editor (February 1, 1975).
   * Microwave Mobile Communications.
   * New York: John Wiley & Sons Inc. ISBN 0-471-43720-4
   */

  // number of path = M
  // x = 1 -> M=6, x = 2 -> M=8, x = 3 -> M=10, x = 4 -> M=12
  int x = m_nbOfPaths->GetValue ();

  for (int i = 0; i < downlinkSubChannels; i++)
    {
      // StartJakes allow us to select a window of 0.5ms into the Jakes realization lasting 3s.
      int startJakes = m_startJakes->GetValue ();

      MultipathForTimeDomain multipathForTimeDomain;

      if (x == 1)
        {
          // SELECTED 6 MULTIPLE PATH FOR JAKES MODEL
          if (speed == 0)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M6_v_0 [j + startJakes]);
                }
            }
          if (speed == 3)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M6_v_3 [j + startJakes]);
                }
            }
          if (speed == 30)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M6_v_30 [j + startJakes]);
                }
            }
          if (speed == 120)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M6_v_120 [j + startJakes]);
                }
            }
        }

      else if (x == 2)
        {
          // SELECTED 6 MULTIPLE PATH FOR JAKES MODEL
          if (speed == 0)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M8_v_0 [j + startJakes]);
                }
            }
          if (speed == 3)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M8_v_3 [j + startJakes]);
                }
            }
          if (speed == 30)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M8_v_30 [j + startJakes]);
                }
            }
          if (speed == 120)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M8_v_120 [j + startJakes]);
                }
            }
        }

      else if (x == 3)
        {
          // SELECTED 6 MULTIPLE PATH FOR JAKES MODEL
          if (speed == 0)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M10_v_0 [j + startJakes]);
                }
            }
          if (speed == 3)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M10_v_3 [j + startJakes]);
                }
            }
          if (speed == 30)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M10_v_30 [j + startJakes]);
                }
            }
          if (speed == 120)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M10_v_120 [j + startJakes]);
                }
            }
        }

      else if (x == 4)
        {
          // SELECTED 6 MULTIPLE PATH FOR JAKES MODEL
          if (speed == 0)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M12_v_0 [j + startJakes]);
                }
            }
          if (speed == 3)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M12_v_3 [j + startJakes]);
                }
            }
          if (speed == 30)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M12_v_30 [j + startJakes]);
                }
            }
          if (speed == 120)
            {
              for (int j = 0; j < 500; j++)
                {
                  multipathForTimeDomain.push_back (multipath_M12_v_120 [j + startJakes]);
                }
            }
        }
      else
        {
          std::cout << " ERROR: Jaks's Model, incorrect M value" << std::endl;
        }


      m_multipath->push_back (multipathForTimeDomain);
    }

  SetLastUpdate ();
}

double
MultipathLossModel::GetValue (int subChannel)
{
  NS_LOG_FUNCTION (this);
  if (NeedForUpdate ())
    {
      SetValue ();
      SetLastUpdate ();
    }

  int now_ms = Simulator::Now ().GetSeconds () * 1000;
  int lastUpdate_ms = GetLastUpdate ().GetSeconds () * 1000;
  int index = now_ms - lastUpdate_ms;

  NS_LOG_FUNCTION (this << subChannel << now_ms
                        << lastUpdate_ms << index << m_multipath->at (subChannel).at (index));

//BEGIN Luca Anchora
//Luca Anchora: I introduce the following line for the multi-BS scenario.
//Since BSs are assigned separate subchannels of the spectrum, their
//list of subchannels does not start from the value 0. I shift the
//value of the parameter subChannel so as the first subchannel in the
//list of the BS corresponds to the index 0.
int firstSubChannel = GetPhy ()->GetDownlinkSubChannels ().at (0);
//END Luca Anchora

  return m_multipath->at (subChannel-firstSubChannel).at (index);
}

//FG

// double
// MultipathLossModel::GetValue (double frequency)
// {
// 
//  double now = Simulator::Now ().GetSeconds ();
//  
// 
//  double pathloss = MultipathLossModel::CalculateValue (frequency,now);
// 
//   return pathloss;
// }

//FG

// double MultipathLossModel::CalculateValue (double frequency, double now)
// 
// {         
//          
//          Ptr<MobilityModel> mobility = 0;
//              
//           
// 	  // this needs to be fixed, we cannot allow a propagation model to need pointers to all PHYs
// 	  mobility = GetPhy ()->GetDownlinkSpectrumPhy ()->GetMobility ()->GetObject<MobilityModel> ();
// 	  Vector speedVector = mobility->GetVelocity ();
//            
// 
// 
// 	  double speed = sqrt (pow (speedVector.x,2) +  pow (speedVector.y,2));
//              
//   
// 	  // number of path = M
// 	  // x = 1 -> M=6, x = 2 -> M=8, x = 3 -> M=10, x = 4 -> M=12
// 	  int x = m_nbOfPaths->GetValue ();
// 	  int m;
// 
// 	  if (x == 1)
// 		  m = 6;
// 	  else if (x == 2)
// 		  m=8;
// 	  else if (x == 3)
// 		  m = 10;
// 	  else if (x == 4)
// 		  m = 12;
//            
//          //////////////////////////////////
//          //static double inc = 0;
//          //inc = inc + 0.001;
//          //now = inc;
//          //std::cout << "time " << now << std::endl;
//         //std::cout << "frequency " << frequency << std::endl;
//         //std::cout << "M " << m << std::endl;
// 
//          /////////////////////////////////
//         
//          
// 	  // calcolo N
// 	  int n =2*((2*m)+1);
// 	  // calcolo la pulsazione della portante
// 	  double wc = 2* PI *frequency;
// 
// 	  int c = 299792458;
// 
// 	  //pulsazione del max shift doppler
// 	  double wm = wc *speed/c ;
// 
// 	  //inizializzaione parte reale e immaginaria risposta del canale
// 	  double hI=0;
// 	  double hQ=0;
//           double hIsum = 0;
//           double hQsum = 0;
//           double hInonsum = 0;
//           double hQnonsum = 0;
//           double betaN = 0;
// 
//          
// 
//       // Calcolo parte reale
//           int count;
// 	  for (count = 1 ; count < (m+1); count++ )
// 	  {
// 		  betaN = (PI*count)/m; //variabile Beta_n
// 		  hIsum = hIsum + (cos(betaN)*cos(cos(2*PI*count/n)*wm*now)); // calcolo sommatoria parte reale
// 	  }
// 
//           hIsum = hIsum*2;
// 	  hInonsum = sqrt(2)*cos(wm*now)*cos(PI/4);     //parte di hI calcolata senza sommatoria
// 	  hI = hIsum + hInonsum;
// 	  hI = hI * (2/sqrt(n));
// 
// 	  //Calcolo parte immaginaria
//           //count = 1;
// 	  for (count = 1; count < m+1; count++ )
// 	  	  {
// 	  		  betaN = (PI*count)/m; //variabile Beta_n
// 	  		  hQsum = hQsum + (sin(betaN)*cos(cos(2*PI*count/n)*wm*now)); // calcolo sommatoria parte immaginaria
// 	  	  }
// 
// 	  	  hQsum = hQsum*2;
// 	  	  hQnonsum = sqrt(2)*cos(wm*now)*cos(PI/4);     //parte di hQ calcolata senza sommatoria
// 	  	  hQ = hQsum + hQnonsum;
// 	  	  hQ = hQ * (2/sqrt(n));
//       
// 	  //Calcolo il modulo
// 	  double pathloss = sqrt(pow(hI,2)+pow(hQ,2)); 
// 	  double pathloss_dB = 10*log10(pathloss);
// 
//           //std::cout << pathloss_dB << " pathloss_dB " << std::endl;	  
// 
// 	return pathloss_dB;
// }



// Added by Francesco Guidolin	
struct complexMultipath
MultipathLossModel::GetValue (double frequency)
{		
	
	if (NeedForUpdate ())
	{
		// compute number of path m
		// x = 1 -> M=6, x = 2 -> M=8, x = 3 -> M=10, x = 4 -> M=12
		int x = m_nbOfPaths->GetValue ();
		
		if (x == 1)
			m_m = 6;
		else if (x == 2)
			m_m=8;
		else if (x == 3)
			m_m = 10;
		else if (x == 4)
			m_m = 12; 
		//NS_LOG_UNCOND ("Update m " << m_m );  
	
		SetLastUpdate ();
	} 
		


	double now = Simulator::Now ().GetSeconds ();
	
	struct complexMultipath multipath = CalculateValue (frequency, now, m_m);
	
	return multipath;
}



struct complexMultipath MultipathLossModel::CalculateValue (double frequency, double now, int m)

{         
	
	Ptr<MobilityModel> mobility = 0;
	
	
	// this needs to be fixed, we cannot allow a propagation model to need pointers to all PHYs
	mobility = GetPhy ()->GetDownlinkSpectrumPhy ()->GetMobility ()->GetObject<MobilityModel> ();
	Vector speedVector = mobility->GetVelocity ();
	
	
	
	double speed = sqrt (pow (speedVector.x,2) +  pow (speedVector.y,2));

	
	int n =2*((2*m)+1);
	double wc = 2* PI *frequency; // pulsation
speed = 8.0;//Luca Anchora, equivalent to 30 Km/h....no mobility!!
	double wm = wc *speed/C ; // max doppler shift
	
	//NS_LOG_UNCOND ("SPEED: " << speed);
	//NS_LOG_UNCOND ("Frequency under analysis: " << frequency);
	//NS_LOG_UNCOND ("Frequency Doppler shift: " << wm/(2*PI));
	
	double hI=0;
	double hQ=0;
	double hIsum = 0;
	double hQsum = 0;
	double hInonsum = 0;
	double hQnonsum = 0;
	double betaN = 0;
		
	
	
	int count;
	for (count = 1 ; count < (m+1); count++ )
	{ 
		betaN = (PI*count)/(m+1); 
		hIsum = hIsum + (cos(betaN)*cos(cos(2*PI*count/n)*wm*now));
	}
	
	hIsum = hIsum*2;
	hInonsum = sqrt(2)*cos(wm*now)*cos(m_betaO);     
	

	hI = (hIsum + hInonsum)*(1/(sqrt(2*m+1)));
			
			
	
	for (count = 1; count < (m+1); count++ )
	{
		betaN = (PI*count)/(m+1); 
		hQsum = hQsum + (sin(betaN)*cos(cos(2*PI*count/n)*wm*now));
	}
	
	hQsum = hQsum*2;
	hQnonsum = sqrt(2)*cos(wm*now)*sin(m_betaO); 
	
	hQ = (hQsum + hQnonsum) * (1/(sqrt(2*m+1)));		
		
	
//	double multipath = sqrt(pow(hI,2)+pow(hQ,2)); 
	//double multipath_dB = 10*log10(multipath);
			
	//NS_LOG_UNCOND ("Associated MULTIPATH value : " << multipath_dB << " [dB]");
	struct complexMultipath mp;
	mp.real = hI;
	mp.im = hQ;
	return mp;
}



} // namespace ns3
