/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2010 IMT Lucca
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Luca Anchora  <luca.anchora@imtlucca.it>
 */


#include "simple-packet-scheduler-modified.h"
#include <ns3/log.h>
#include <ns3/pointer.h>
#include "ue-net-device.h"
#include "enb-net-device.h"
#include "simple-packet-scheduler.h"
#include "enb-net-device.h"
#include "rlc-entity.h"
#include "rrc-entity.h"
#include "lte-mac-header.h"
#include "ns3/mac48-address.h"
#include "enb-mac-entity.h"
#include "ue-manager.h"
#include "ue-record.h"
#include "amc-module.h"
#include <ns3/random-variable.h>
#include <math.h>
#include <ns3/simulator.h>
#include <stdio.h>


NS_LOG_COMPONENT_DEFINE ("SimplePacketSchedulerModified");

namespace ns3 {


NS_OBJECT_ENSURE_REGISTERED (SimplePacketSchedulerModified);

#define PKT_SIZE 100
#define ASCII_0 (0x30)

struct optimum_list_element{
	Mac48Address user;
	unsigned int subc;
	int subc_state;
};

SimplePacketSchedulerModified::SimplePacketSchedulerModified ()
{
	NS_LOG_FUNCTION (this);
	SetDevice (0);

	attempted_list_ = 0;
	curr_throughput_ = 0;
	total_throughput_ = 0;
	tot_pkts_tx_ = 0;
	curr_pkts_tx_ = 0;
	not_configured = 1;
	init_subframes = 0;
}


/** 
 * Configuration of the main object properties. We prefer to do it in a
 * separate method instead of inside the constructor.
 */
void SimplePacketSchedulerModified::config (Ptr<EnbNetDevice> enb)
{
	SetDevice (enb);

	attempted_list_ = new List();
	optimum_list_ = 0;
	curr_throughput_ = 0;
	total_throughput_ = 0;
	tot_pkts_tx_ = 0;
	curr_pkts_tx_ = 0;
	num_ue_ = enb->GetUeManager ()->GetNRegisteredUes();
	quantum_ = 10000;
	Creq_ = 0;
	curr_max_throughput_ = 0;
	tot_max_throughput_ = 0;

	queues_ = 0;
	queues_ = (struct flow_queue*)malloc( sizeof(struct flow_queue) * num_ue_ );
	if(!queues_){
		printf("Not enough space for allocating flow queues in the scheduler. Terminating the process\n");
		exit(1);
	}
	cumulative_credits_ = 0;
	for(int i=0; i<num_ue_; i++)
	{
		queues_[i].priority = 0; //maximum priority to all flows as a default setting
		queues_[i].credit_counter = 0.0;
		queues_[i].pkt_queue = new List();
		queues_[i].byte_tx = new circular_queue();
		queues_[i].curr_byte_tx = 0.0;
		queues_[i].tot_byte_tx = 0.0;
	}

	not_configured = 0;
}

TypeId
SimplePacketSchedulerModified::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::SimplePacketSchedulerModified")
    .SetParent<SimplePacketScheduler> ()
    .AddConstructor<SimplePacketSchedulerModified> ()
  ;
  return tid;
}


/**
 * In the destructor we deallocate all the pointers that have been
 * dynamically allocated, in order to avoid memory leaks.
 */
SimplePacketSchedulerModified::~SimplePacketSchedulerModified ()
{
	NS_LOG_FUNCTION (this);
	if(queues_){
		for(int i=0; i<num_ue_; i++)
		{
			if(queues_[i].pkt_queue)
				delete queues_[i].pkt_queue;
			queues_[i].pkt_queue = 0;

			if(queues_[i].byte_tx)
				delete queues_[i].byte_tx;
			queues_[i].byte_tx = 0;
		}
		free(queues_);
	}
	queues_ = 0;

	if(attempted_list_)
		delete(attempted_list_);
	attempted_list_ = 0;

	if(optimum_list_)
		delete(optimum_list_);
	optimum_list_ = 0;
}


/**
 * Calculation of the Nash Bargaining function in correspondence of the argument D.
 * For our system we have NBF(D)=F(D)*T(D), where F(D) and T(D) are fairness and
 * throughput for Cmax=D.
 */
double SimplePacketSchedulerModified::NB_function (double D)
{
	double fairn_D=0, throughput_D=0;

	//We have to simulate an allocation to measure the resulting
	//fairness and throughput
	struct flow_queue* copy_queues = (struct flow_queue*)malloc( sizeof(struct flow_queue) * num_ue_ );
	fakefillIn_list(D, copy_queues);//virtual scheduling done to compute the fairness and throughput that
	                                //would result by choosing Cmax = D.

	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	int num_downlink_subch = enb->GetPhy()->GetDownlinkSubChannels().size();

	//In the following iteration we construct the resource mapping 
	//<user, subchannel>. Packets are selected among those present in
	//the attempted list.
	char subc_allocated[num_downlink_subch];//array to mark the subchannels that have already been allocated
	memset(subc_allocated, 0, num_downlink_subch);
	
	double curr_th = 0.0;
	int tot_pkts_tx = tot_pkts_tx_;
	int curr_pkts_tx = curr_pkts_tx_;
	throughput_D = total_throughput_;
	build_optimum_list(num_ue_, num_downlink_subch);

	optimum_list_->start_scanning();
	struct optimum_list_element* tmp = (struct optimum_list_element*)optimum_list_->next_item();
	Packet* pkt = 0;
	LteMacHeader header;
	Mac48Address destinationMac;
	int count_pkts = 0; //counter of the number of packets selected for the transmission
	while(count_pkts < Creq_ && tmp)
	{
		while(tmp && subc_allocated[tmp->subc])//only subchannels not allocated are considered
			tmp = (struct optimum_list_element*)optimum_list_->next_item();

		if(!tmp)
			break;

		attempted_list_->start_scanning();

		//Look up tmp->user in the attempted list, and when it is found
		//then the correspondent pkt is transmitted
		do
		{
			pkt = (Packet*)attempted_list_->next_item();
			if(!pkt) break;
			pkt->RemoveHeader(header);
			destinationMac = header.GetDestination ();
			pkt->AddHeader(header);
		} while(pkt && tmp->user != destinationMac);

		if(pkt)
		{ //match found
			count_pkts++;
			subc_allocated[tmp->subc] = 1;

			//We calculate the throughput using Shannon's formula
			double spec_eff = GetMacEntity ()->GetAmcModule ()->GetSpectralEfficiencyFromCQI(tmp->subc_state);
			double SINR_nat = log(5.0 * 0.00005 )*(1 - pow(2, spec_eff))/1.5;
			curr_th += spec_eff*180000*log2(1 + SINR_nat);
			attempted_list_->remove_item((void*)pkt);
			if(pkt){
				delete pkt;
				pkt = 0;
			}
			tot_pkts_tx++;
			curr_pkts_tx++;
		}
		
		tmp = (struct optimum_list_element*)optimum_list_->next_item();
	}

	if(count_pkts < Creq_)//some resources are still free
	{
		attempted_list_->start_scanning();
		int i = 0;
		Packet* p;
		while(count_pkts < Creq_)
		{
			p = 0;
			while(subc_allocated[i])
				i++;
			
			p = (Packet*)attempted_list_->next_item();

			count_pkts++;
			subc_allocated[i] = 1;

			//We calculate the throughput using the Shannon's formula
			double spec_eff = GetMacEntity ()->GetAmcModule ()->GetSpectralEfficiencyFromCQI(15);
			double SINR_nat = log(5.0 * 0.00005 )*(1 - pow(2, spec_eff))/1.5;
			curr_th += spec_eff*180000*log2(1 + SINR_nat);
			attempted_list_->remove_item((void*)p);
			if(p)
			{
				delete p;
				p = 0;
			}
			tot_pkts_tx++;
			curr_pkts_tx++;
		}

	}

	throughput_D += curr_th;

	delete optimum_list_;
	optimum_list_ = 0;

	//Compute the fairness that would result from this allocation
	double sum = 0;
	double sum2 = 0;

	for(int i=0; i<num_ue_; i++)
	{
		sum += copy_queues[i].tot_byte_tx;
		sum2 += (copy_queues[i].tot_byte_tx * copy_queues[i].tot_byte_tx);
	}

	if(sum2!=0)
		fairn_D = (sum*sum)/(num_ue_ * sum2);
	else
		fairn_D = 1.0;

	//Resulting throughput from this virtual allocation
	double BW = num_downlink_subch * 180000*5.55; //Hz * max_spectral_efficiency
	double normal_th = throughput_D/(tot_pkts_tx * BW);//normalization


	if(copy_queues)
		free(copy_queues);
	copy_queues = 0;

	return fairn_D * normal_th;
}


/**
 *  As the function fillIn_list it fills in the attempted_list_. But it is just a trial
 *  that is done to calculate the values of fairness and throughput that would result
 *  by using this particular length for the attempted_list_. In other words, this is
 *  a virtual scheduling, no packet will be transmitted in the end. Since this operation
 *  should be virtual, it should not modify the state of the flow queues, that's why
 *  we work on a copy of them.
 */
void SimplePacketSchedulerModified::fakefillIn_list(int Cmax, struct flow_queue* copy_queues)
{
	attempted_list_->empty();
	Packet* p;
	double cumulative_credits = 0;

	for(int i=0; i<num_ue_; i++)
	{
		copy_queues[i].priority = queues_[i].priority; //maximum priority to all flows as a default setting
		copy_queues[i].tot_byte_tx = queues_[i].tot_byte_tx;
	}

	//Increment the credit counter for each flow queue.
	for(int i=0; i<num_ue_; i++)
	{
		copy_queues[i].credit_counter = queues_[i].credit_counter + quantum_;
		copy_queues[i].curr_byte_tx = 0.0;
		cumulative_credits += copy_queues[i].credit_counter;
	}

	//For each flow choose the corresponding number of packets to be transmitted
	//according to its level of credit.
	int pkt_selected = 0;
	int portion = 0;//transmission opportunity given to each flow
	int count_bytes = 0;
	UniformVariable* u = new UniformVariable();
	int count_rounds = 0;
	char dest_MAC_address[12] = {'1', ':', '2', ':', '3', ':', '4', ':', '5', ':', '6', 0};//dumb MAC address
	Mac48Address* destinationMac;
	LteMacHeader mac_header;

	for(int i=1; i<=num_ue_; i++)
	{
		count_rounds++;
		portion = (int)((copy_queues[i-1].credit_counter * Cmax)/cumulative_credits);
		pkt_selected += portion;
		if(pkt_selected > Cmax)
			portion = Cmax - (pkt_selected - portion);

		count_bytes = 0;
		for(int j=0; j<portion; j++)
		{
			p = new Packet(PKT_SIZE);
			dest_MAC_address[10] = i + ASCII_0;
			destinationMac = new Mac48Address(dest_MAC_address);
			mac_header.SetDestination (*destinationMac);
			p->AddHeader (mac_header);
			attempted_list_->enqueue(p);
			copy_queues[i-1].credit_counter -= PKT_SIZE;
			copy_queues[i-1].curr_byte_tx += PKT_SIZE;
			copy_queues[i-1].tot_byte_tx += PKT_SIZE;
		}
	}

	if(pkt_selected!=Cmax)
	{//There are still some pkts left because of integer truncation.
	 //They are chosen randomly from the queues.
		unsigned int r;
		for(; pkt_selected!=Cmax; pkt_selected++)
		{
			r = u->GetInteger(1.0, num_ue_);
			p = new Packet(PKT_SIZE);
			dest_MAC_address[10] = r + ASCII_0;
			destinationMac = new Mac48Address(dest_MAC_address);
			mac_header.SetDestination (*destinationMac);
			p->AddHeader (mac_header);
			attempted_list_->enqueue(p);
			copy_queues[r-1].credit_counter -= PKT_SIZE;
			copy_queues[r-1].curr_byte_tx += PKT_SIZE;
			copy_queues[r-1].tot_byte_tx += PKT_SIZE;

		}
	}

	delete u;
}


/**
 * Game theoretic algorithm to compute the value fo Cmax. The value returned is the
 * point of maximum of the Nash function, which is estimated in a numerical way
 * by exploiting the properties of its derivative.
*/
int SimplePacketSchedulerModified::determine_Cmax()
{
	int min = Creq_;
	int max = num_ue_ * Creq_;
	int I = max - min;
	double epsilon = 0.1;
	double D0, D1, D2;
	double f0, f1, f2;
	double sl1, sl2;
	double dist;

	while(I > epsilon)
	{
		dist = 1.0/4.0 * I;
		D0 = min + dist;
		D1 = D0 + dist;
		D2 = D1 + dist;
		f0 = NB_function(D0);
		f1 = NB_function(D1);
		f2 = NB_function(D2);
		sl1 = (f1-f0)/dist;
		sl2 = (f2-f1)/dist;
		if(sl1<=0){
			max = D1;
		}else{
			min = D0;
			if(sl2 > 0)
				min = D1;
			else
				max = D2;
		}
		
		I = max - min;
	}
	
	return (int)(min + I/2.0);
}


/**
 * The packet scheduler prepares the list of candidate packets for the transmission.
 */
void SimplePacketSchedulerModified::fillIn_list(int Cmax)
{
	attempted_list_->empty();
	Packet* p;
	cumulative_credits_ = 0;
	//Increment the credit counter for each flow queue.
	for(int i=0; i<num_ue_; i++)
	{
		queues_[i].credit_counter += quantum_;
		queues_[i].curr_byte_tx = 0.0;
		cumulative_credits_ += queues_[i].credit_counter;
	}

	//For each flow choose the corresponding number of packets to be transmitted
	//according to its level of credit.
	int pkt_selected = 0;
	int portion = 0;//transmission opportunity given to each flow
	int count_bytes = 0;
	UniformVariable* u = new UniformVariable();
//	int first = (int)u->GetInteger(0.0, num_ue_);
	int count_rounds = 0;
	char dest_MAC_address[12] = {'1', ':', '2', ':', '3', ':', '4', ':', '5', ':', '6', 0};//dumb MAC address
	Mac48Address* destinationMac;
	LteMacHeader mac_header;

	for(int i=1; i<=num_ue_; i++)
	{
		count_rounds++;
		portion = (int)((queues_[i-1].credit_counter * Cmax)/cumulative_credits_);
		pkt_selected += portion;
		if(pkt_selected > Cmax)
			portion = Cmax - (pkt_selected - portion);

		count_bytes = 0;
		for(int j=0; j<portion; j++)
		{
			p = new Packet(PKT_SIZE);
			dest_MAC_address[10] = i + ASCII_0;
			destinationMac = new Mac48Address(dest_MAC_address);
			mac_header.SetDestination (*destinationMac);
			p->AddHeader (mac_header);
			attempted_list_->enqueue(p);
			queues_[i-1].credit_counter -= PKT_SIZE;
			queues_[i-1].curr_byte_tx += PKT_SIZE;
			queues_[i-1].tot_byte_tx += PKT_SIZE;
		}
		queues_[i-1].byte_tx->enqueue(queues_[i-1].curr_byte_tx);
	}

	if(pkt_selected!=Cmax)
	{//There are still some pkts left because of integer truncation.
	 //They are chosen randomly from the queues.
		unsigned int r;
		for(; pkt_selected!=Cmax; pkt_selected++)
		{
			r = u->GetInteger(1.0, num_ue_);
			p = new Packet(PKT_SIZE);
			dest_MAC_address[10] = r + ASCII_0;
			destinationMac = new Mac48Address(dest_MAC_address);
			mac_header.SetDestination (*destinationMac);
			p->AddHeader (mac_header);
			attempted_list_->enqueue(p);
			queues_[r-1].credit_counter -= PKT_SIZE;
			queues_[r-1].byte_tx->enqueue(PKT_SIZE);
			queues_[r-1].curr_byte_tx += PKT_SIZE;
			queues_[r-1].tot_byte_tx += PKT_SIZE;
		}
	}

	delete u;
}


/**
 *  Building of the ordered list of tuples <user, subchannel, subchann_state>. The list is 
 *  ordered with respect to the value of the subchannel state. In this way, we have a
 *  clear and ordered list of the channel state for each user that can be exploited
 *  by the resource allocator for the transmission of packets on the best subchannels.
 */
void SimplePacketSchedulerModified::build_optimum_list(int tot_users, int tot_subc)
{
	optimum_list_ = new List();
	struct optimum_list_element* element;
	struct optimum_list_element* tmp;
	int position_counter = 0;
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	Ptr<UeRecord> ueRecord;
	UeRecord::CqiFeedbacks feedbacks;
	Mac48Address* destinationMac;
	char dest_MAC_address[12] = {'1', ':', '2', ':', '3', ':', '4', ':', '5', ':', '6', 0};

	for(int user = 0; user< tot_users; user++)
	{
		dest_MAC_address[10] = user + 1 + ASCII_0;
		destinationMac = new Mac48Address(dest_MAC_address);
		ueRecord = enb->GetUeManager ()->GetUeRecord (*destinationMac);
		if(!ueRecord)//no feedback from this UE
			continue;

		feedbacks = ueRecord->GetCqiFeedbacks ();

		int totSubChannels = feedbacks.size ();
		for (int i = 0; i < totSubChannels; i++)
		{
			int subChannel = feedbacks.at (i).m_subChannelId;
			int cqi = feedbacks.at (i).m_cqi;
		
			element = (struct optimum_list_element*)malloc(sizeof(struct optimum_list_element));
			element->user = *destinationMac;
			element->subc = subChannel;
			element->subc_state = cqi;

			//Insert the element in the correct position into the ordered list according to
			//the value for the subchannel state. The list is scanned and when the correct
			//position is found, the element is inserted.
			//Elements are put in decreasing order of CQI.
			optimum_list_->start_scanning();
			tmp = (struct optimum_list_element*)optimum_list_->next_item();
			position_counter = 1;
			while(tmp!=0)
			{
				if(tmp->subc_state < element->subc_state)
				{
					break;
				}else{
					tmp = (struct optimum_list_element*)optimum_list_->next_item();
					position_counter++;
				}
			}//end while(tmp!=0)

			optimum_list_->put_position(element, position_counter);

		} //end for(subc=0; subc<tot_subc; subc++)
	} //end for(user = 0; user< tot_users; user++)

	return;
}


void
SimplePacketSchedulerModified::DoRunPacketScheduler (void)
{
	NS_LOG_FUNCTION (this);
	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	if(not_configured){
		config(enb);
		not_configured = 0;
	}

	Ptr<PdcchMapIdealControlMessage> msg = CreateObject<PdcchMapIdealControlMessage> ();// Create PDCCH message
	double mcs;
	int num_downlink_subch = enb->GetPhy()->GetDownlinkSubChannels().size();
	Creq_ = num_downlink_subch; //cell load
	Ptr<Packet> p;
	Ptr<PacketBurst> pb = CreateObject<PacketBurst> ();
	LteMacHeader header;
	Mac48Address destinationMac;

	if(init_subframes<2)
	{
		// The eNB Sends a virtual packet in order to allow UEs to compute the SINR
		Ptr<PacketBurst> pb = CreateObject<PacketBurst> ();
		Ptr<Packet> p = new Packet (1);
		
		Mac48Address from = Mac48Address::ConvertFrom (enb->GetAddress ());
		Mac48Address to = Mac48Address::ConvertFrom (enb->GetAddress ());
		
		LteMacHeader header;
		header.SetSource (from);
		header.SetDestination (to);
		p->AddHeader (header);
		
		pb->AddPacket (p);
		enb->SetPacketToSend (pb);
		enb->StartTransmission ();

		init_subframes++;
		return;
	}

	//read the value of Cmax from an external configuration file
	int Cmax; 
	FILE* fp_inputCmax = fopen("val_D.txt", "r");
	fscanf(fp_inputCmax, "%d", &Cmax);
	fclose(fp_inputCmax);
	if(Cmax==0)
		Cmax = determine_Cmax();//automatic calculation of the value of Cmax

	fillIn_list(Cmax);
	build_optimum_list(num_ue_, num_downlink_subch);
	max_throughput();

	//In the following iteration we construct the resource mapping 
	//<user, subchannel>. Packets are selected among those present in
	//the attempted list.
	char subc_allocated[num_downlink_subch];//array to mark the subchannels that have already been allocated
	memset(subc_allocated, 0, num_downlink_subch);
	
	curr_throughput_ = 0.0;
	curr_pkts_tx_ = 0;

	optimum_list_->start_scanning();
	struct optimum_list_element* tmp = (struct optimum_list_element*)optimum_list_->next_item();
	Packet* pkt = 0;
	
	int count_pkts = 0; //counter of the number of packets selected for the transmission
	while(count_pkts < Creq_ && tmp)
	{
		while(tmp && subc_allocated[tmp->subc])//only subchannels not allocated are considered
			tmp = (struct optimum_list_element*)optimum_list_->next_item();

		if(!tmp)
			break;

		attempted_list_->start_scanning();
		
		//Look for the user in the attempted list, and when it is found
		//then the correspondent pkt is transmitted
		do
		{
			pkt = (Packet*)attempted_list_->next_item();
			if(!pkt) break;
			pkt->RemoveHeader(header);
			destinationMac = header.GetDestination ();
			pkt->AddHeader(header);
		} while(pkt && tmp->user != destinationMac);

		if(pkt)
		{ //match found
			count_pkts++;
			subc_allocated[tmp->subc] = 1;
			Mac48Address from = Mac48Address::ConvertFrom (enb->GetAddress ());
			pkt->RemoveHeader(header);
			header.SetSource(from);
			pkt->AddHeader(header);
			pb->AddPacket (pkt->Copy());
			mcs = GetMacEntity ()->GetAmcModule ()->GetMCSFromCQI (tmp->subc_state);
			Ptr<UeRecord> uer = enb->GetUeManager ()->GetUeRecord (destinationMac);
			msg->AddNewRecord (PdcchMapIdealControlMessage::DOWNLINK,
                                 tmp->subc,
                                 uer->GetUe ()->GetObject<UeNetDevice> (),
                                 mcs);

			//We calculate the throughput using Shannon's formula
			double spec_eff = GetMacEntity ()->GetAmcModule ()->GetSpectralEfficiencyFromCQI(tmp->subc_state);
			double SINR_nat = log(5.0 * 0.00005 )*(1 - pow(2, spec_eff))/1.5;
			curr_throughput_ += spec_eff*180000*log2(1 + SINR_nat);
			attempted_list_->remove_item((void*)pkt);
			if(pkt)
			{
				delete pkt;
				pkt = 0;
			}
			tot_pkts_tx_++;
			curr_pkts_tx_++;
		}
		 
		tmp = (struct optimum_list_element*)optimum_list_->next_item();
	}

	//Some subchannels might still be free
	if(count_pkts < Creq_)
	{
		attempted_list_->start_scanning();
		int i = 0;
		Packet* p = 0;
		while(count_pkts < Creq_)
		{
			p = 0;
			while(subc_allocated[i])
				i++;
			
			p = (Packet*)attempted_list_->next_item();

			count_pkts++;
			subc_allocated[i] = 1;
			Mac48Address from = Mac48Address::ConvertFrom (enb->GetAddress ());
			p->RemoveHeader(header);
			header.SetSource(from);
			destinationMac = header.GetDestination();
			p->AddHeader(header);
			pb->AddPacket (p->Copy());

			mcs = GetMacEntity ()->GetAmcModule ()->GetMCSFromCQI (15);
			msg->AddNewRecord (PdcchMapIdealControlMessage::DOWNLINK,
                                 i,
                                 enb->GetUeManager ()->GetUeRecord (destinationMac)->GetUe ()->GetObject<UeNetDevice> (),
                                 mcs);

			//We calculate the throughput by using Shannon's formula
			double spec_eff = GetMacEntity ()->GetAmcModule ()->GetSpectralEfficiencyFromCQI(15);
			double SINR_nat = log(5.0 * 0.00005 )*(1 - pow(2, spec_eff))/1.5;
			curr_throughput_ += spec_eff*180000*log2(1 + SINR_nat);
			attempted_list_->remove_item((void*)p);
			if(p)
			{
				delete p;
				p = 0;
			}
			tot_pkts_tx_++;
			curr_pkts_tx_++;
		}

	}


	total_throughput_ += curr_throughput_;

	delete optimum_list_;
	optimum_list_ = 0;

	//Manage the credits for the packets in the attempted list that have not been transmitted
	attempted_list_->start_scanning();
	pkt = (Packet*)attempted_list_->next_item();
	uint8_t buff[6];
	Mac48Address addrMAC;
	while(pkt)
	{
		pkt->RemoveHeader(header);
		addrMAC = header.GetDestination();
		addrMAC.CopyTo(buff);
		queues_[buff[5]-1].credit_counter += PKT_SIZE;
		queues_[buff[5]-1].byte_tx->reduce_last_sample(PKT_SIZE);
		queues_[buff[5]-1].curr_byte_tx -= PKT_SIZE;
		attempted_list_->remove_item((void*)pkt);
		if(pkt)
		{
			delete pkt;
			pkt = 0;
		}
		pkt = (Packet*)attempted_list_->next_item();
	}

	//Start the transmission
	enb->SetPacketToSend (pb);
	GetMacEntity ()->GetObject<EnbMacEntity> ()->SendPdcchMapIdealControlMessage (msg);
	enb->StartTransmission ();

	log_values();//call the log function to trace the current value of the performance indices
}


/**
 *  Calculate the max throughput achievable with the tranmsission  of Creq_
 *  packets according to the current channel state, regardless of the fairness.
 *  In other words, given the current channel state, we calculate the max
 *  throughput we could achieve by transmitting on the best Creq_ subchannels.
 */
void SimplePacketSchedulerModified::max_throughput()
{
	if(!optimum_list_)
		return;

	Ptr<EnbNetDevice> enb = GetMacEntity ()->GetDevice ()->GetObject<EnbNetDevice> ();
	int num_downlink_subch = enb->GetPhy()->GetDownlinkSubChannels().size();
	char subc_allocated[num_downlink_subch];
	memset(subc_allocated, 0, num_downlink_subch);
	curr_max_throughput_ = 0;

	optimum_list_->start_scanning();
	struct optimum_list_element* tmp = (struct optimum_list_element*)optimum_list_->next_item();
	
	int count_pkts = 0; //counter of the number of packets selected for the transmission
	while(count_pkts < Creq_ && tmp)
	{
		while(tmp && subc_allocated[tmp->subc])//only subchannels not allocated are considered
			tmp = (struct optimum_list_element*)optimum_list_->next_item();

		if(!tmp)
			break;

		count_pkts++;
		double spec_eff = GetMacEntity ()->GetAmcModule ()->GetSpectralEfficiencyFromCQI(tmp->subc_state);
		double SINR_nat = log(5.0 * 0.00005 )*(1 - pow(2, spec_eff))/1.5;
		curr_max_throughput_ += spec_eff*180000*log2(1 + SINR_nat);
	}

	tot_max_throughput_ += curr_max_throughput_;

}


void SimplePacketSchedulerModified::log_values()
{
	FILE* log_file = fopen("sched_rra_log.txt", "a");
	
	double th_norm = throughput_norm();
	double fairn = fairness_norm();
	fprintf(log_file, "%f\t%f\t%f\n", Simulator::Now().GetSeconds(), fairn, th_norm);

	fclose(log_file);
}

double SimplePacketSchedulerModified::throughput_norm()
{
	if(tot_max_throughput_)
		return (total_throughput_ / tot_max_throughput_);
	else
		return 0;

}

/**
 * Jain's index is used to measure the fairness.
 */
double SimplePacketSchedulerModified::fairness_norm()
{
	//Compute the fairness that would result from this allocation
	double sum = 0;
	double sum2 = 0;

	for(int i=0; i<num_ue_; i++)
	{
		sum += queues_[i].tot_byte_tx;
		sum2 += (queues_[i].tot_byte_tx * queues_[i].tot_byte_tx);
	}

	if(sum2!=0)
		return ((sum*sum)/(num_ue_ * sum2));
	else
		return 1.0;

}

}//namespace