/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2011 IMT Lucca
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Luca Anchora  <luca.anchora@imtlucca.it>
 */

#ifndef MULTIOP_PACKET_SCHEDULER_H
#define MULTIOP_PACKET_SCHEDULER_H

#include "packet-scheduler.h"
#include "ns3/mac48-address.h"
#include "ns3/packet-burst.h"
#include "List.h"
#include "miscellaneous.h"

namespace ns3 {

class List;

typedef struct flow_queue prioQueue;

struct optimum_list_element{
	Mac48Address user;
	unsigned int subc;
	int subc_state;
};

typedef struct{
	int subchannelID;
	char allocated;
	char priority;
	Mac48Address user;
	int CQI;
} alloc_map_element;

/**
 * This class implements a simple credit-based packet scheduler. It also
 * implements the resource allocator: when a sequence of packets has been
 * selected for the transmission, it decides the mapping with the subchannels
 * (the resources that can be used) according to an efficiency goal. In
 * particular, it tries to exploit the multiuser diversity for maximizing
 * total throughput.
 *
 * The communication between the packet scheduler and the resource allocator is
 * done through a list of candidate packets: the scheduler fills it in with a 
 * number of possible packets (Cmax), and the resource allocator selects a subset
 * of them (Creq) on the basis of the resources available and on the desired 
 * channel load.
 */

class MultiOpPacketScheduler : public PacketScheduler
{

public:
	MultiOpPacketScheduler ();
	virtual ~MultiOpPacketScheduler ();

	static TypeId GetTypeId (void);

	virtual void DoRunPacketScheduler (void);
	virtual void TxVirtualPacket ();
	virtual void* NewAlloc(void* e, List* l);
	virtual void DoTx(void* e);

private:
	/**
	 * \brief Allocate a flow queue per each UE registered to the eNB.
	 * \param num_ue number of users registered to the eNB.
	 * \param enb pointer to the net device.
	*/
	void allocate_queues(int num_ue, Ptr<EnbNetDevice> enb);

	void build_optimum_list(void);
	void build_optimum_list_noshar(void);
	void log_values();

void printChannelState(); //for debugging reasons

private:
	prioQueue* m_queues; //flow queues. A priority queue is associated to each flow/user.
	Ptr<PacketBurst> m_txList; //list of packets to be transmitted
	List* optimum_list_;
	List* optimum_list_noshar_;
	double m_totThroughput;
	double m_currThroughput;
	double m_totCapacity;
	double m_currCapacity;
	int m_countTx; //the allocation map is updated after 100 transmissions (100 ms)
	alloc_map_element* prop_alloc_map;
	alloc_map_element* alloc_map_noshar;
	int m_unallocatedSlots;
	double m_noSharCap;
	double m_noSharThroughput;
	double m_initialSharCap;
	double m_initialSharThroughput;
	int m_commonSlotsUsed;
};


} //namespace

#endif /* MULTIOP_PACKET_SCHEDULER_H */
